
try:
  import numpy as np
except ImportError:
  print "Numpy not installed"

try:
  from PIL import Image, ImageDraw, ImageFont
except ImportError:
  print "Numpy not installed"

def square_crop(im, target_size=None):
  """ Crop image to `target_size`. If that's None the image is squared 
  to the smallest size
  """
  w = im.size[0]
  h = im.size[1]

  target_size = target_size if target_size else min(w, h)

  dx = (w - target_size) / 2
  dy = (h - target_size) / 2

  return im.crop((dx, dy, dx + target_size, dy + target_size))

def draw_time(im, font, time=None, dump=None, padding_x=0.25, padding_y=0.07):

  padding = padding_x
  w, h = im.size

  y_loc = h - padding_y * h

  draw = ImageDraw.Draw(im)

  if dump:
    line = "%04i" % dump
    text_w, text_h = draw.textsize(line, font=font)

    x_loc = padding * w / 2.0
    draw.text((x_loc, y_loc), line, (255,255,255), font=font)

  if time:
    line = "%0.f min" % time
    text_w, text_h = draw.textsize(line, font=font)

    x_loc = w - text_w - padding * w / 2.0
    draw.text((x_loc, y_loc), line, (255,255,255), font=font)

def color_to_alpha(im_in, color=(0, 0, 0, 255), threshold=50):

    im_out = im_in.convert("RGBA")
    pixdata = im_out.load()

    for y in xrange(im_in.size[1]):
        for x in xrange(im_in.size[0]):
            pd = pixdata[x, y]
            
            if pixdata[x, y] == color:
                pixdata[x, y] = (0, 0, 0, 0)
            else:               
                avg = int((pd[0] + pd[2] + pd[1]) / 3.0)
                
                if avg < threshold:
                    pixdata[x, y] = (pd[0], pd[1], pd[2], avg)
                           
                
    return im_out   

def alpha_composite(src, dst):
    '''
    Return the alpha composite of src and dst.

    Parameters:
    src -- PIL RGBA Image object
    dst -- PIL RGBA Image object

    The algorithm comes from http://en.wikipedia.org/wiki/Alpha_compositing
    '''
    
    # http://stackoverflow.com/a/3375291/190597
    # http://stackoverflow.com/a/9166671/190597
    
    src = np.asarray(src)
    dst = np.asarray(dst)
    out = np.empty(src.shape, dtype = 'float')
    alpha = np.index_exp[:, :, 3:]
    rgb = np.index_exp[:, :, :3]
    src_a = src[alpha]/255.0
    dst_a = dst[alpha]/255.0
    out[alpha] = src_a+dst_a*(1-src_a)
    old_setting = np.seterr(invalid = 'ignore')
    out[rgb] = (src[rgb]*src_a + dst[rgb]*dst_a*(1-src_a))/out[alpha]
    np.seterr(**old_setting)    
    out[alpha] *= 255
    np.clip(out,0,255)
    # astype('uint8') maps np.nan (and np.inf) to 0
    out = out.astype('uint8')
    out = Image.fromarray(out, 'RGBA')
    return out





#!/usr/bin/env python

"""
Layers the fv and enuc together
"""


import argparse
import os
import re
import subprocess
import tempfile

from PIL import Image, ImageDraw, ImageFont

from lcse_tools import image_util as iu

#temp_template = '/u/sciteam/sandalsk/scratch/timed/%s'

def make_stuff(prefix_a, prefix_b, output_path, output_prefix, dump_range):
  """ Layer images with `prefix_b` on top of `prefix_a`
  """

  # We want to combine images from A on top of B
  mask_a = prefix_a + "%04d.png"
  mask_b = prefix_b + "%04d.png"

#  rtp = lcse.rtplot_reader(project_path)
#  dumps = rtp.dump_map.keys()
#  dumps.sort()
  dumps = range(dump_range[0], dump_range[1] + 1)

  print "Processing dummps %s" % dumps

  path = os.path.join(os.path.abspath(output_path), output_prefix)

  if not os.path.exists(path):
     os.makedirs(path)

  for dump in dumps:
        
    filename_out = os.path.join(path, '%s-%04d.png' % (output_prefix, dump))

    if os.path.exists(filename_out):
      continue
    
    print filename_out
    
    # load fv image
    try:
        # Open
        image_a = Image.open(mask_a % dump).convert("RGBA")
        image_b = Image.open(mask_b % dump).convert("RGBA")
        
        # Crop
        image_a = iu.square_crop(image_a, image_a.size[1])
        image_b = iu.square_crop(image_b, image_b.size[1])
        
        # Make the second image transparent
        image_b = iu.color_to_alpha(image_b, threshold=30)      
        image_a = iu.alpha_composite(image_b, image_a)
        
#        draw_time(image_a, font, dump=dump, time=rtp.dump_map[dump]['T'])
        image_a.save(filename_out)
        
    except IOError as e:
        print e
        continue

def build_argparse():

  parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  parser.add_argument('--top', help='Top image', required=True)
  parser.add_argument('--bottom', help='Bottom image', required=True)
  parser.add_argument('--output-prefix', help='Prefix for output filenames <prefix>-<dump>.png', required=True)
  parser.add_argument('--output-path', help='Path to store image files', default=".")
  parser.add_argument('--dumps', nargs='+',type=int, help='Dump range to process (e.g. --dumps 10 200)')

  return parser
  
def main():

  parser = build_argparse()
  args = parser.parse_args()

  if  len(args.dumps) < 2:
    print "Specifiy dump range like --dumps 10 200"
    return

  make_stuff(args.bottom, args.top, args.output_path, args.output_prefix, args.dumps)


  return

  parser.add_argument('--add-time', help='Add time from an RTplot.ppm file')
  parser.add_argument('--font', help='/home/stou/lcse_tools/fonts/Roboto-Bold.ttf')

  prefixes = args.prefixes if args.prefixes else os.listdir(args.i)

  return

if __name__ == '__main__':
  main()



#!/usr/bin/env python

"""
Create movies out of a series of pngs. Add dump and/or time value from an RTplot.ppm file.


Running 

.. code:: bash

  $ pngs2images.py

in a directory containing directories of images (e.g. ./foo/foo_0001.png,foo_0002.png,..,foo_1020.png, ; ./bar/bar_0001.png, ..., bar_1929.png) will
produce a movie for each directory encountered (foo_0001-1020.avi, bar_0001-1929.avi)
"""

import argparse
import os
import pkg_resources
import re
import subprocess
import tempfile

from PIL import Image, ImageDraw, ImageFont

import lcse
from lcse_tools import image_util as iu



def filter_files(path, prefix=None, suffix=None):
  # Switch to https://docs.python.org/2/library/fnmatch.html

  try:
    cond = lambda x : (x.startswith(prefix) if prefix else True) and (x.endswith(suffix) if suffix else True)
    files = [os.path.join(path, f) for f in os.listdir(path) if cond(f)]
    return files
  except Exception as e:
    print e


# 1. Create temp directory to work in
# 2. Fully composite image
# 3. Write out to temp file (don't overwrite if it exists)
# 4. Make movie

def add_time_to_pngs(image_path, prefix, font_name, temp_prefix, time_map={}, font_size=50):

  print temp_prefix
  print font_name

  temp_path = os.path.join(temp_prefix, prefix)

  if not os.path.exists(temp_path):
    os.makedirs(temp_path)

  font = ImageFont.truetype(font_name, font_size)

  image_files = filter_files(image_path, suffix='.png')
  image_files.sort()

  dump_re = re.compile('(.*)([-_])([\d]{4,6})(.*)')

  prefix = os.path.basename(os.path.commonprefix(image_files))

  for filename in image_files:

    m = dump_re.match(filename)

    if not m:
      continue

    dump = int(m.group(3))
    filename_out = os.path.join(temp_path, os.path.basename(filename))

    if os.path.exists(filename_out):
      continue

    im = Image.open(filename)
    iu.draw_time(im, font, dump=dump, time=time_map.get(dump))

    print filename, '->', filename_out

    im.save(filename_out)

  return temp_path

def make_movie_from_pngs(image_path, **kwargs):
  prefix = kwargs.get('prefix', '')

  add_time = kwargs.get('add_time')

  if add_time is not None:
    if add_time:
      rtp = lcse.rtplot_reader(add_time[0])
      time_map = dict((dump, val['T']) for dump, val in rtp.dump_map.items())
    else:
      time_map = {}

    image_path = add_time_to_pngs(image_path, prefix, kwargs.get('font'), kwargs.get('tmp_dir'), time_map,
                                  font_size=kwargs.get("font_size"))

  fps = kwargs.get('fps', 24)
  tune = kwargs.get('tune', 'stillimage')
  threads = kwargs.get('threads') or 'auto'

  print prefix

  files = filter_files(image_path, prefix=prefix, suffix='.png')
  dumps = [int(f[-8:-4]) for f in files]
  dumps.sort()

  if not dumps:
    print "Nothing found for %s in %s" % (prefix, image_path)
    return

  dmin, dmax = min(dumps), max(dumps)

  elements = []
  if kwargs.get('video_prefix'): elements.append(kwargs.get('video_prefix'))
  elements.append(prefix)
  elements.append('%04i-%04i' % (dmin, dmax))
  if kwargs.get('video_suffix'): elements.append(kwargs.get('video_suffix'))
  elements.append(str(fps))

  video_filename = "%s.avi" % ('_'.join(elements))

#  video_suffix = "%s_%s" % (kwargs.get('video_suffix'), fps if kwargs.get('video_suffix') else str(fps)
#  video_filename = '%s_%04i-%04i_%s.avi' % (video_prefix, dmin, dmax, video_suffix)
  video_path = os.path.abspath(os.path.join(kwargs.get('movies_path', '.'), video_filename))

  if not os.path.exists(kwargs.get('movies_path', '.')):
    os.makedirs(kwargs.get('movies_path', '.'))

  if os.path.exists(video_path):
    print "File %s exists, skipping" % video_path
    return

  if kwargs.get('process_only'):
    print "Not making movie %s. Processing only" % video_path
    return

  # Make Video
  args = ['mencoder', 'mf://%s*.png' % prefix,
          '-mf', 'fps=%i:type=png' % fps,
          '-oac', 'copy',
          '-ovc', 'x264', '-x264encopts',
          'preset=veryslow:tune=%s:crf=15:frameref=15:fast_pskip=0:threads=%s' % (tune, threads),
          '-o', video_path]

  print("Output file is %s" % video_path)

  p = subprocess.Popen(args, cwd=image_path)
  s, i = p.communicate()
  p.wait()

def build_argparse():

  default_font = os.path.abspath(os.path.join(pkg_resources.resource_filename(__name__, '../../fonts'), 'Roboto-Bold.ttf'))


  parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

  parser.add_argument('-i', help='Path to images base directory',  default='.')
  parser.add_argument('-o', help='Path to video save directory', default='.')
  parser.add_argument('--fps', nargs='+',type=int, help='List of fps to create', default=[18])
  parser.add_argument('--prefixes', nargs='+', help='List of prefixes to try. Otherwise everything inside input-path is used')
  parser.add_argument('--video-prefix', help='Prefix to add to output video files')
  parser.add_argument('--video-suffix', help='Suffix to add to output video files')

  parser.add_argument('--add-time', nargs='*', help='Specify RTplot.ppm file to use for timing info')
  parser.add_argument('--font', help='Font to use', default=default_font)
  parser.add_argument('--font-size', type=int, help='Size of font to use', default=50)
  parser.add_argument('--tmp-dir', help='Temporary directory to store intermediate files', default='/tmp/timed')
  parser.add_argument('--threads', help='Number of threads to use for video encoding')
  parser.add_argument('--process-only', help='Process images only, do not make a movie', action='store_true')

  return parser
  

def main():

  parser = build_argparse()
  args = parser.parse_args()

  if args.prefixes:
    prefixes = args.prefixes
  else:
    prefixes = os.listdir(args.i)
    prefixes.sort()

  print "Using prefixes", prefixes, args.i, args.o

  for fps in args.fps:
    for prefix in prefixes:
      try:

        in_path = os.path.join(args.i, prefix)

        if not os.path.isdir(in_path):
          continue

        make_movie_from_pngs(in_path, prefix=prefix,
                             video_prefix=args.video_prefix or '',
                             video_suffix=args.video_suffix or '',
                             fps=fps, movies_path=args.o, font=args.font,
                             add_time=args.add_time, tmp_dir=args.tmp_dir,
                             font_size=args.font_size, threads=args.threads,
                             process_only=args.process_only)
      except Exception as e:
        print "Problem", e

  return

if __name__ == '__main__':
  main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
C Pre-Processor #define evaluator.

Usage:

.. code:: bash

  eval_pp.py input_file.F -o output_file.F -D israyprofile=0 isrestart=0
'''

__author__ = "Stou Sandalski"
__email__ = "stou@icapsid.net"
__date__ = '25 Apr 2015'
__license__ = "Apache 2.0"
__version__ = "1.0"


import argparse
import logging
import re

# TODO: Remove blank statements

class def_stmt(object):
  
  def __init__(self, l):
    self.line = l
    
    tmp = l.split()
    self.name = tmp[1]
    self.expr = tmp[2:]

  def __str__(self):
    return str(self.line)

class if_stmt(object):
  
  def __init__(self, l):
    self.line = l
    self.first = []
    self.second = []

    tmp = l.split()

    self.expr = ' '.join(tmp[1:])
    self.eval_res = None
    self.prefix = tmp[0]

    self._ifdef = tmp[0] == '#ifdef'
    
  def __str__(self):

    out = [self.line]
    out += [str(s) for s in self.first]

    if self.second:
      out += ['#else\n']
      out += [str(s) for s in self.second]

    out += ['#endif\n']
    s = ''.join(out)

    return s

  def eval_cond(self, symbols):
    ''' Evaluate this if statment '''

    # Propagate the eval to nested if statements
    first = [s.eval_cond(symbols)[1] if isinstance(s, if_stmt) else s for s in self.first]
    second = [s.eval_cond(symbols)[1] if isinstance(s, if_stmt) else s for s in self.second]

    expr = self.expr.strip()

    res = None

    if self._ifdef:
      # We don't want to remove undefined symbols at this point. 
      # Could be an option in the future.
      res = True if expr in symbols else None
    else:
      expr = filter_expr(expr, symbols)

      try:
      	res = bool(eval(expr))
      except NameError:
        res = None
      except SyntaxError as e:
        print expr, self.expr

    res_str = ''

    if res is True:
      res_str = ''.join(str(s) for s in first)
    elif res is False:
      res_str = ''.join(str(s) for s in second)
    else:
      out = ['%s %s\n' % (self.prefix, expr)]
      out += [str(s) for s in first]

      if second:
        out += ['#else\n']
        out += [str(s) for s in second]

      out += ['#endif\n']
      res_str = ''.join(out)
   
    return res, res_str

def filter_expr(expr, symbols):
  '''Tokenize expr and replace symbols in the stream. Otherwise doing a search 
     and replace on a symbol can partially match other symbols (isSPU vs isSPUC)
  '''
  res_expr = []
    
  for s in re.split('(\W)', expr):
    if len(s) > 0:
      res_expr.append(s if s not in symbols else symbols[s])

  return ''.join(res_expr)

def process_file(file_in, file_out, **kwargs):

  # Read the file in full (we have enough ram)
  f = open(file_in)
  lines = f.readlines() 
  f.close()

  ix = 0

  eval_list = {'ismpi':1}

  symbols = {}

  statements_raw, ix = consume_lines(lines, symbols, ix, 0)

  eval_symbols = kwargs.get('symbols')

  statements = []

  if eval_symbols:
    for s in statements_raw:
      if isinstance(s, if_stmt):
         r, res_s = s.eval_cond(eval_symbols)
         statements.append(res_s)
      else:
        statements.append(s)
  else:
    statements = statements_raw

  f = open(file_out, 'w')
  f.write(''.join(str(s) for s in statements))
  f.close()

def consume_lines(lines, symbols, ix, level):
  
  statements = []

  lines_ct = len(lines)

  ifs = None

  while ix < lines_ct:

    top = lines[ix]

    if top.startswith('#define'):
      ds = def_stmt(top)

      if level == 0:
        symbols[ds.name] = ds

      statements.append(ds)
      ix += 1
    elif top.startswith('#if'):
      ifs = if_stmt(top)
      ifs.first, ix = consume_lines(lines, symbols, ix+1, level+1)
    elif top.startswith('#else'):
      if ifs:
        ifs.second, ix = consume_lines(lines, symbols, ix+1, level+1)
      else:
        return statements, ix
    elif top.startswith('#endif'):
      if ifs:
        statements.append(ifs)
        ix += 1
        ifs = None
      else:
        return statements, ix
    else:
      statements.append(top)
      ix += 1

  return statements, ix
  
def build_argparse():

  parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

  parser.add_argument('file_in', metavar='input', help='Intput file')
  parser.add_argument('-o', dest='file_out', metavar='output', default='code.F', help='Output file')
  parser.add_argument('-D', dest='symbols', metavar='key=value', nargs='*', help='Space separated key=value to evaluate')
  
  return parser

def main():

  parser = build_argparse()

  args = parser.parse_args()

  symbols = dict(s.split('=') for s in args.symbols) if args.symbols else {}

  process_file(args.file_in, args.file_out, symbols=symbols)

if __name__ == '__main__':
  main()



#!/bin/env python

"""
Check for missing files in directories containing sequences of files. For example sequences of rendered images or plots, .hv files, etc.

Usage:

.. code:: bash

  check_missing.py <path>
"""

__author__ = 'Stou Sandalski <stou@icapsid.net>'
__license__= 'Public Domain'



import logging
import os
import re

#TODO: Add a total

# create file handler which logs debug messages
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

try:
  from colorlog import ColoredFormatter

  formatter = ColoredFormatter(
  # "%(log_color)s%(levelname)-8s%(reset)s %(blue)s%(message)s",
  #  "%(log_color)s %(asctime)s - %(name)s - %(levelname)s : %(white)s%(message)s",
    "%(log_color)s %(asctime)s - %(levelname)s : %(white)s%(message)s",
    datefmt=None,
    reset=True,
    log_colors={
      'DEBUG':    'cyan',
      'INFO':     'green',
      'WARNING':  'yellow',
      'ERROR':    'red',
      'CRITICAL': 'red,bg_white',
    },
    secondary_log_colors={},
    style='%'
    )
  ch.setFormatter(formatter)
except:
#  formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
  formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
  ch.setFormatter(formatter)

log = logging.getLogger('check_missing')
log.addHandler(ch)
log.setLevel(logging.INFO)
#log.propagate = False

fmt_missing = 'Missing \033[1;33m {0:>3d}\033[1;m in {1:28}: {2} in range {3:04d}-{4:04d}'
fmt_ok = '{0:25} range {1:04d}-{2:04d}: \033[1;32m ok \033[1;m'

def check_missing_items(path):
  log.debug("Checking for missing items in %s " % path)

  dump_re = re.compile('(.*)-([\d]{4})\.')

  contents = [os.path.join(path, f) for f in os.listdir(path)]
  files = [os.path.basename(p) for p in contents if os.path.isfile(p)]

  dump_maps = {}

  for f in files:
    m = dump_re.match(f)

    if not m:
      continue

    prefix, dump = m.groups()

    dm = dump_maps.get(prefix, [])
    dm.append(int(dump))
    dump_maps[prefix] = dm

  for name, items in dump_maps.items():

    items.sort()

    items_theory = range(items[0], items[-1] + 1)
    items = set(items)

    missing = list(set(items_theory).difference(items))
    missing.sort()

    missing_ct = len(missing)

    if missing_ct <= 0:
      log.info(fmt_ok.format(name, items_theory[0], items_theory[-1]))
      return []

    log.warning(fmt_missing.format(missing_ct, name, ', '.join(str(d) for d in missing), items_theory[0], items_theory[-1]))

  dirs = [p for p in contents if os.path.isdir(p)]
  dirs.sort()

  for d in dirs:
    check_missing_items(d)


if __name__ == '__main__':

  import sys

  if len(sys.argv) < 2:
    print "Usage check_missing.py <path>"
  else:
    path = sys.argv[1]
    check_missing_items(path)



#!/usr/bin/env python

'''
This file contains plotting routines for RProfile data. It is for now
the repository of all plot functions... should be moved somewhere.

# TODO: Add an installable script we can just run
'''

import os
import sys
import numpy as np
import logging

from matplotlib import colors
from matplotlib.path import Path
import matplotlib.patches as patches

import lcse
import image_util

formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
ch = logging.StreamHandler()
ch.setFormatter(formatter)

log = logging.getLogger(__name__)
log.propagate = False
log.addHandler(ch)
log.setLevel(logging.INFO)
#log.setLevel(logging.ERROR)

#########################################
#                                       #
#  Basic Building Blocks for Plotting   #
#                                       #
#########################################

colors = ['g', 'b', 'y', 'm']
styles = ['o-', 'v-', 'D-', 'x-', '<-', '*-', 'H-','+-', '^-','>-', '+-']

default_plot_path = './plots'

def get_styles(count):
  """Return `count` number of style attributes for plotting"""

  return [styles[ i % len(styles)] for i in range(count)]

def new_figure(figsize=None):
  """ 
     Create a new figure of size `figsize` that has a white background 
  """

#  log.info("Creating new figure")

  figsize = figsize if figsize else (12, 8)

  if "matplotlib" not in sys.modules:
    import matplotlib
    matplotlib.use("Agg")
   
  import matplotlib.pyplot as plt

  fig = plt.figure(figsize=figsize, facecolor='w', edgecolor='k')
  fig.patch.set_facecolor('white')
  fig.patch.set_alpha(1.0)
  return fig

def update_axis(ax, rp=None, **kwargs):
  """ `xlim` and `ylim` """

  xlim = kwargs.get('xlim')
  ylim = kwargs.get('ylim')

  ax.legend(loc=0)
  ax.grid(True)

  if xlim:
    ax.set_xlim(xlim)
  elif rp:
    ax.set_xlim((rp.get('radinner'), rp.get('radouter')))

  if ylim:
    ax.set_ylim(ylim)

def get_filename(prefix, **kwargs):
  """ 
  Construct a filename from `dump`, `bucket`, `path`, `filename`.
  """

  dump = kwargs.get('dump')
  bucket = kwargs.get('bucket', 0)
  filename = kwargs.get('filename')

  path = os.path.join(kwargs.get('path', default_plot_path), prefix)

  if not filename:
    pieces = [prefix]

    if bucket: pieces.append("_%03i" % bucket)
    
    if dump: pieces.append("_%04i" % dump)
     
    pieces.append('.png')
    
    filename = ''.join(pieces)

  path_out = os.path.join(path, filename)
  log.debug("get_filename %s" % path_out)
  return path_out

def check_exists(prefix, **kwargs):

  filename = get_filename(prefix, **kwargs)

  if os.path.exists(filename):
    log.info("Skipping figure %s because it exists" % filename)
    return True
  else:
    return False

def save_fig(fig, prefix, **kwargs):
  """ 
  Save a figure 

  Keywords:

  `dump` 

  `dpi` the dpi for the figure, default is 200

  `path`

  `filename`

  `bucket`

  `tight` True/False to use fig.tight_layout

  """

  dpi = kwargs.get('dpi', 200)

  filename = get_filename(prefix, **kwargs)
  path = os.path.dirname(filename)

  if not os.path.exists(path):
    os.makedirs(path)

  if os.path.exists(filename):
    log.info("Skipping figure %s because it exists" % filename)
    return

  if kwargs.get("tight"):
    fig.tight_layout()

  fig.savefig(filename, dpi=dpi)
  log.info("Saved figure to %s" % filename)
  
  
#################################################
#                                               #
#  Generic or multi-purpose plotting routines   #
#                                               #
#################################################

def plot_var_one_bucket(rp, var, fig, **kwargs):
  ''' Plot the `var` with variable

  `xlim` and `ylim` are the axis limits
  '''

  bucket = kwargs.get('bucket', 0)
  dump = rp.get('dump')

  prefix = kwargs.get('prefix', '%s' % var)

  if check_exists(prefix, dump=dump, **kwargs):
    return

  y = rp.get('y_hi' if var == 'fv_hi' else 'y')

  data = rp.get_table(var)
  var_avg = data[0,:,bucket]
  var_min = data[1,:,bucket]
  var_max = data[2,:,bucket]
  var_sd = data[3,:,bucket]

  ax = fig.add_subplot(111)

  bucket_name = 'bucket %i' % bucket if bucket > 0 else 'global bucket'
  ax.set_title('%s for %s dump %04i' % (var, bucket_name, dump))
  ax.set_xlabel('Radius ($10^3 km$)')

  ax.plot(y, var_avg, 'k')
#  ax.plot(y, var_min,'b', label="min")
#  ax.plot(y, var_max,'r', label="max")
#  ax.errorbar(y, var_avg, yerr=var_sd,
#              label='average',
#              elinewidth=2, fmt='o', ecolor='g')

  update_axis(ax, rp=rp)

  save_fig(fig, prefix, dump=dump, **kwargs)
  return ax

def plot_log_var_one_bucket(rp, var, fig, **kwargs):
  ''' Plot the `var` multiplied by the volume of that shell
  `xlim` and `ylim` are the axis limits
  '''

  bucket = kwargs.get('bucket', 0)
  dump = rp.get('dump')

  prefix = kwargs.get('prefix', 'log_%s' % var)

  if check_exists(prefix, dump=dump, **kwargs):
    return

  y = rp.get('y_hi' if var == 'fv_hi' else 'y')
  d = rp.get_table(var)[0,:,bucket]

  ax = fig.add_subplot(111)

  bucket_name = 'bucket %i' % bucket if bucket > 0 else 'global bucket'

  ax.set_title('%s for %s dump %04i' % (var, bucket_name, dump))
  ax.set_xlabel('Radius ($10^3 km$)')
  ax.plot(y, np.log10(d),'k')

  update_axis(ax, rp=rp, **kwargs)

  save_fig(fig, prefix, dump=dump, **kwargs)
  return ax

def plot_var_many_buckets(rp, var, fig, **kwargs):
  ''' Create a plot of several buckets against the global
  of a single variable. Only min/max/sd for the global are
  shown

  `buckets` if specified is a list of buckets to add to the plot
            otherwise we plot everything
  '''

#  fig = kwargs.get('fig', new_figure())
  buckets = kwargs.get('buckets', range(rp.bucket_count + 1))

#  if check_exists(prefix, dump=dump, **kwargs):
#    return

  y = rp.get('y_hi' if var == 'fv_hi' else 'y')
  d = rp.get_table(var)

  ax = fig.add_subplot(111)

  ax.set_title('%s for many bucket' % (var))
  ax.set_xlabel('Radius ($10^3 km$)')

  ax.plot(y, d[0,:,0],'k')
  ax.plot(y, d[1,:,0],'b', label="min")
  ax.plot(y, d[2,:,0],'r', label="max")
  ax.errorbar(y, d[0,:,0], yerr=d[3,:,0],
                    elinewidth=2, fmt='o', ecolor='g')

  for bucket in buckets:
    ax.plot(y, d[0,:,bucket],'k')

  update_axis(ax, rp=rp, **kwargs)
  return ax

def plot_log_var_many_buckets(rp, var, fig, **kwargs):
  ''' Create a plot of several buckets against the global
  of a single variable. Only min/max/sd for the global are
  shown

  `buckets` if specified is a list of buckets to add to the plot
            otherwise we plot everything

  `min`

  `max`

  `labeled` if True each bucket plot is labled individualy 
  '''

  buckets = kwargs.get('buckets', range(rp.bucket_count + 1))

#  if check_exists(prefix, dump=dump, **kwargs):
#    return

  y = rp.get('y_hi' if var == 'fv_hi' else 'y')
  d = rp.get_table(var)

  ax = fig.add_subplot(111)

  ax.set_title('%s for many bucket' % (var))
  ax.set_xlabel('Radius ($10^3 km$)')
  ax.set_ylabel('$\log_{10}(%s)$' % var)

  ax.plot(y, np.log10(d[1,:,0]), 'b', label="min")
  ax.plot(y, np.log10(d[2,:,0]), 'r', label="max")
  ax.plot(y, np.log10(d[0,:,0]))

  labeled = kwargs.get('labeled')

  if labeled:
    for i, s in zip(buckets, style):
      ax.plot(np.log10(data[0,:,i]), s, label="%i" % i)
  else:
    for bucket in buckets:
      ax.plot(y, np.log10(d[0,:,bucket]), 'k')

  update_axis(ax, rp=rp, **kwargs)
  return ax

def plot_two_log_vars_buckets(rp, var1, var2, fig, buckets, xlim=None):
    
  data1 = rp.get_table(var1)
  data2 = rp.get_table(var2)

  styles = get_styles(len(buckets))

  ax = fig.add_subplot(111)
  ax2 = ax.twinx()
  
  print data1[0,:,5].shape
  
  for i, b in enumerate(buckets):
      ax.plot(np.log10(data1[0,:,i]), styles[i], label="%s - %i" % (var1, b))
      ax2.plot(np.log10(data2[0,:,i]), label="%s - %i" % (var2, b))

  ax2.plot(np.log10(data1[0,:,0]), "o", label="global %s" % var2)   
  ax.set_title("dump %i" % rp.get("dump"))
     
  if xlim:
      ax.set_xlim(xlim)
#     ax.set_ylim([-8, 0])
#     ax2.set_ylim([-8, 0])
 
  ax.grid()
  ax.legend(loc=2)
#    ax2.legend(loc=1)
  return ax, ax2
  
  
  
##############################################
#                                            #
#  Specialized and Single Purpose Routines   #
#                                            #
##############################################

def plot_ceul_mach_global(rp, fig, **kwargs):

  prefix = 'ceul_mach_global'

  dump = rp.get('dump')

  if check_exists(prefix, dump=dump, **kwargs):
    return

  y = rp.get('y')
  ceul = rp.get('ceul')
  mach = rp.get('mach')

  ax = fig.add_subplot(111)
  ax.set_title('$C_{eul}$ and Mach number for dump %4i' % dump)
  ax.set_xlabel('Radius ($10^3 km$)')

  ax.plot(y, ceul[2], '-', label='$\max C_{eul}$')
  ax.plot(y, ceul[0], '-', label='$C_{eul}$')
  ax.plot(y, ceul[1], '-', label='$\min C_{eul}$')

  ax2 = ax.twinx()
  ax2.plot(y, mach[2], 'r', label='max(M)')
  ax2.plot(y, mach[0], '+k', label='M')
  ax2.plot(y, mach[1], 'g', label='min(M)')

  update_axis(ax, rp=rp)

  ax.legend(loc=2, prop={'size':6})
  ax2.legend(prop={'size':6})

  save_fig(fig, prefix, dump=dump, **kwargs)
  return ax

def plot_ekr_ekt_entropy(rp, fig, **kwargs):
  ''' Do not use this method to loop over buckets and plot many of them
  since it will be quite inefficient.
  '''

  prefix = 'ekr_ekt_entropy_global'

  dump = rp.get('dump')

  if check_exists(prefix, dump=dump, **kwargs):
    return

  y = rp.get('y')
  ekr = rp.get('ekr')
  ekt = rp.get('ekt')
  p = rp.get('p')
  rho = rp.get('rho')

  entropy = p[0, :] / (rho[0, :]**(5./3.))

  ix_entropy = entropy > 0.0
  log_entropy = np.log10(entropy[ix_entropy])

  fig.clear()

  ax = fig.add_subplot(111)
  ax.set_title('ekr/ekt and entropy for dump %04i' % (dump))
  ax.set_xlabel('Radius ($10^3 km$)')

  ax.plot(y, ekr[0, :], 'r', label='ekr')
  ax.plot(y, ekt[0, :], 'g', label='ekt')

  ax2 = ax.twinx()
  ax2.plot(y[ix_entropy], log_entropy, 'b', label='log(A)')

  update_axis(ax, rp=rp)
  ax.legend(loc=2)
  ax2.legend(loc=1)

  save_fig(fig, prefix, dump=dump, **kwargs)
  return ax

def plot_unwraped_quantity(rp, data, label, fig, **kwargs):
  """ Plot a scatter plot where the points represent bucket centers
  
  `rp` ray profile instance
  
  `data` data with dimension of the size of the buckets.

  `label` the text string to use for the plot quantities
  
  `fig` instances
  
  
  """

  path = kwargs.get('path', '.')
  save = kwargs.get('save', False)
  normalize = kwargs.get('normalize', True)
  scale = kwargs.get('scale', 3000.0)
  global_range = kwargs.get('global_range')
  
  dump = rp.get('dump')
  centers = rp.get_centers()

  phi_x = centers[0]
  theta_y = centers[1]
    
#     if not isinstance(scales, list):
#         scales = [scales for i in range(len(variables))]
    
#    colors = range(rp.get('nbuckets'))
#    cm = plt.cm.get_cmap('YlOrRd')
  cm = plt.cm.get_cmap('bwr')
  
  mm = (data.min(), data.max()) if global_range is None else global_range
  
  if normalize:
    data = (data - mm[0]) / (mm[1] - mm[0])
    mm = (0.0, 1.0)

  colors = data
      
  fig.clear()
  ax = fig.add_subplot(111)
  
  ax.set_title('%s dump %4i' % (label, dump))
  ax.set_xlabel(r'$\phi$')
  ax.set_ylabel(r'$\theta$')
  ax.set_ylim([0.0, np.pi])

#    sc = ax.scatter(phi_x, theta_y, s=scale * data, c=colors, 
#                    alpha=0.5, vmin=mm[0], vmax=mm[1], cmap=cm)

  sc = ax.scatter(phi_x, theta_y,  s= scale * data,
                  alpha=0.5, vmin=mm[0], vmax=mm[1], norm=True)
    
  ax.set_xticks(np.pi * np.arange(-1, 1.5, 0.5))
  ax.set_xticklabels(["-$\pi$", "-$\pi/2$","0", "$\pi/2$", "$\pi$"])
  
  ax.set_yticks(np.pi * np.arange(0, 1.25, 0.25))
  ax.set_yticklabels(["0", "$\pi/4$", "$\pi/2$", "$3 \pi/4$", "$\pi$"])
     
  if kwargs.get("annotate"):
    for i in range(rp.get('nbuckets')):
      ax.annotate(str(i + 1), (phi_x[i], theta_y[i]))

  x = range(phi_x)
  ax.xticks(x, labels)

#    plt.colorbar(sc)

  if save:
    save_fig(fig, label, dump=dump)

  return ax

def plot_bucket_map(rp, fig=None):
  """Plot a map of bucket centers"""

  centers = rp.get_centers()
  phi_x = centers[0]
  theta_y = centers[1]
  
  if fig is None:    
    fig = new_figure()
  else:
    fig.clear()
    
  ax = fig.add_subplot(111)

  ax.set_title("Bucket Map")
  ax.set_xlabel(r'$\phi$')
  ax.set_ylabel(r'$\theta$')
  ax.set_ylim([0.0, np.pi])
      
  sc = ax.scatter(phi_x, theta_y)
     
  ax.set_xticks(np.pi * np.arange(-1, 1.5, 0.5))
  ax.set_xticklabels(["-$\pi$", "-$\pi/2$","0", "$\pi/2$", "$\pi$"])
  
  ax.set_yticks(np.pi * np.arange(0, 1.25, 0.25))
  ax.set_yticklabels(["0", "$\pi/4$", "$\pi/2$", "$3 \pi/4$", "$\pi$"])

  for i in range(rp.get('nbuckets')):
     ax.annotate(str(i + 1), (phi_x[i], theta_y[i]))

  ax.grid()

  return ax
  
def plot_unwrapped_map(rp, fig=None, buckets=[], even_stretch=False, fill_colors=[], highlight=[], highlight_color=None):
  """
  `rp` ray profile
  `fig` figure to plot this on
  `buckets` is a list of bucket IDs (1 based)
  `even_stretch` takes the cos of phi to spread the centers out more
  `fill_color` is an array of colors
  `highlight` the given bucket IDs (1-based bucket number) will be highlighted with `highlight_color`
  """

  centers = -rp.get_centers()[2:,]
  theta_center = np.arctan2(centers[1,:], centers[0,:])
  phi_center = np.arccos(centers[2,:])

  if not buckets:
    buckets = range(centers.shape[1])

  if not fig:
    fig = new_figure()
    
  if highlight and not fill_colors:
    highlight_color = highlight_color or 'indigo'
    fill_colors = [highlight_color if i+1 in highlight else 'white' for i in buckets]

  #def get_corners(rp):
  normals = rp.get_normals()
  corners = np.zeros((3, 3, normals.shape[2]))
  corners[:,0,:] = np.cross(normals[:,0,:], normals[:,1,:], axis=0)
  corners[:,1,:] = np.cross(normals[:,1,:], normals[:,2,:], axis=0)
  corners[:,2,:] = np.cross(normals[:,2,:], normals[:,0,:], axis=0)  
  corners[:,:,:] /= np.sqrt((corners * corners).sum(0))

  theta = np.arctan2(corners[1,:,:], corners[0,:,:]).reshape((1, 3, -1))
  phi = np.arccos(corners[2,:]).reshape((1, 3, -1))
  
  corners_rad = np.vstack((theta, phi))

  for i in range(theta_center.shape[0]):
    positive = 0
    negative = 0
    ambigous = []
    
    for j in range(3):
        
      if (corners_rad[1,j,i] < 1e-7) or (corners_rad[1,j,i] >= np.pi - 1e-7):
        corners_rad[0,j,i] = theta_center[i]
      
      if (abs(theta_center[i]) >= np.pi - 1e-7):
        if theta_center[i] > 0.0 and corners_rad[0,j,i] < 0.0:
          corners_rad[0,j,i] += 2 * np.pi
  
      if (abs(corners_rad[0,j,i]) >= np.pi - 1e-7):
        ambigous.append(j)

      if corners_rad[0,j,i] > 0.0:
        positive += 1
      else:
        negative += 1

    if positive > 0 and negative > 0:
      sign = -1 if theta_center[i] < 0.0 else 1

      for a in ambigous:
        corners_rad[0,a,i] = sign * abs(corners_rad[0,a,i])
      
  if even_stretch:
    corners_rad[1,:,:] = -np.cos(corners_rad[1,:,:])

  codes = [Path.MOVETO,
           Path.LINETO,
           Path.LINETO,
           Path.CLOSEPOLY,
           ]

  ax = fig.add_subplot(111)
  
  if fill_colors:
      get_patch = lambda path, ix: patches.PathPatch(path, facecolor=fill_colors[ix], lw=1)
  else:
      get_patch = lambda path, ix: patches.PathPatch(path, facecolor='white', lw=1)

  for b in buckets:
    verts = corners_rad[:,:,b].T.tolist() + [[0., 0.]]
    path = Path(verts, codes)
    patch = get_patch(path, b)
    ax.add_patch(patch)

  ax.set_xlim(-1.01 * np.pi, 1.19 * np.pi)
  ax.set_ylim(0, np.pi)
  ax.grid(True)
   
  ax.set_xticks(np.pi * np.arange(-1, 1.5, 0.5))
  ax.set_xticklabels(["-$\pi$", "-$\pi/2$","0", "$\pi/2$", "$\pi$"])

  ax.set_yticks(np.pi * np.arange(0, 1.25, 0.25))
  ax.set_yticklabels(["0", "$\pi/4$", "$\pi/2$", "$3 \pi/4$", "$\pi$"])

  for i in range(rp.get('nbuckets')):
    ax.annotate(str(i+1), (theta_center[i], phi_center[i]))
  
  return ax


def plot_energy(data_path, label, fig, stride=1, style="b-"):
  rp_set = lcse.rprofile_set(path=data_path, stride=stride)

  volumes = rp_set.get_dump().get_cell_volumes()

  t = []
  energy = []

  for rp in rp_set:
    dump = rp.get('dump')

    timerescaled = rp.get('timerescaled')
    t.append(timerescaled)

    enuc = rp.get_table('enuc')[0,:,1:]
    enuc *= volumes

    energy.append(enuc.sum())

  ax = fig.add_subplot(1,1,1)
  ax.ticklabel_format(useOffset=False)
  ax.grid(True)
  ax.set_ylabel("energy")
  ax.set_xlabel("time [min]")

  ax.plot(t, log(energy), style, label=label)

#  if save:
#    save_fig(fig, label)
  return ax

def plot_diff_from_avg(r, rp_table, filter_fn=None, fig=None):
  """`rp_table` is a table with the same format as returned by .get_table().
  
  pass array through `filter_fn`, example filter_fn=np.log10 will make the plot
  log
  """
  
#     if check_exists(prefix, dump=dump, **kwargs):
#         return
  
  diffs = rp_table[0,:,1:] - rp_table[0,:,0].reshape((-1, 1))
  diff_size = sqrt((diffs * diffs).sum(1))

  if not fig:
    fig = new_figure()

  ax = fig.add_subplot(111)
      
  if filter_fn:
    ax.plot(r, filter_fn(diff_size))
  else:
    ax.plot(r, diff_size)
      
  ax.grid()
  return ax

from matplotlib.patches import FancyArrowPatch
from mpl_toolkits.mplot3d import proj3d

class RayProfileCenter(FancyArrowPatch):
  def __init__(self, row, *args, **kwargs):
    FancyArrowPatch.__init__(self, (0,0), (0,0), *args, **kwargs)
    self.center = row

  def draw(self, renderer):
    xs, ys, zs = proj3d.proj_transform(self.center[0], self.center[1], self.center[2], renderer.M)
    self.set_positions((0, 0),(xs, ys))
    FancyArrowPatch.draw(self, renderer)

def plot_centers_3d(rp, buckets, fig=None, **kwrds):
  """ Make a 3d plot of the bucket centers
  
  `buckets` list of buckets to plot 
  `fig` a figure object
  """
  
  from mpl_toolkits.mplot3d import Axes3D
  from itertools import product, combinations

  if not fig:
    fig = new_figure(figsize=(10, 10))

  ax = fig.gca(projection='3d')
#  ax.set_aspect("equal")
  size = 2

  # Draw cube
  r = [-size, size]
  for s, e in combinations(np.array(list(product(r,r,r))), 2):
    if np.sum(np.abs(s-e)) == r[1]-r[0]:
      ax.view_init(elev=90., azim=-90.0)
      ax.plot3D(*zip(s,e), color="b")

  ax.set_xlabel('X axis')
  ax.set_ylabel('Y axis')
  ax.set_zlabel('Z axis')

  # Draw axis
#  a = RayProfileCenter((1,0,0), mutation_scale=20, lw=1, arrowstyle="-|>", color="k")
#  a = RayProfileCenter((0,1,0), mutation_scale=20, lw=1, arrowstyle="-|>", color="k")
#  ax.add_artist(a)

  ax.add_artist(ax.text(0.0, -2.0, -2.0, '10', text='X'))
  ax.add_artist(ax.text(-2.0, 0.0, -2.0, '10', text='Y'))
  ax.add_artist(ax.text(-2.0, -2.0, 0.0, '10', text='Z'))

  # Draw a point
  ax.scatter([0],[0],[0],color="g",s=100)
    
  rp_centers = rp.get_centers()

  for i in range(1, rp_centers.shape[1] + 1):
    if i not in buckets:
      continue
    
    center = size * rp_centers[2:,i-1]

    a = RayProfileCenter(center, mutation_scale=20, lw=1, arrowstyle="-|>", color="k")
    ax.add_artist(a)
    t = ax.text(1.15 * center[0], 1.15 * center[1], 1.15 * center[2], '10', text='%i' % i , color='k')
        
    ax.add_artist(t)

  ax.grid()

  return ax

def plot_centers_on_image(rp, buckets=None, img=None, fig=None, index_x=0, index_y=1, scale=1.15, alpha=0.6, prefix=None):
    """
    
    `buckets` a list of buckets to plot

    `img` can be an image which will be used in `imshow(img)` or a string in which
    case the image will be loaded, or none.
    
    `index_x` and `index_y` are the coordinate index to draw as x and as y. By default
    these are index_x = 0, index_y = 1 which gives the standard coordinates. Setting this to 0, 2 will
    provide a top view
    
    `scale` is the size of the arrows.
    
    `alpha` background image alpha
    
    `fig` figure instance
    
    """
        
    index_z = tuple(set((0, 1, 2)).difference((index_x, index_y)))[0]
    axis_labels = ['X', 'Y', 'Z']
    
    if not fig:
      fig = new_figure()
    
    ax = fig.add_subplot(111)
    
    if img:
        # Draw the image
        if isinstance(img, str):
          img = Image.open(img)

        img = image_util.square_crop(img)
        ax.imshow(img, zorder=0, extent=[-1.0, 1.0, -1.0, 1.0], alpha=alpha)

    # Draw the buckets
    if buckets:
      bucket_ids = [i - 1 for i in buckets]
      centers = scale * rp.get_centers()[2:,bucket_ids]
    else:
      centers = scale * rp.get_centers()[2:,]
      buckets = range(centers.shape[1])
        
    x = np.zeros(centers.shape[1])

    q = ax.quiver(x, x, centers[index_x,:], centers[index_y,:], angles='xy', scale_units='xy', scale=1.0)

#     center_colors = np.zeros(centers.shape[1])
#     center_colors[centers[index_z,:] > 0.0] = 1.0
#     q = ax.quiver(x, x, centers[index_x,:], centers[index_y,:], center_colors, angles='xy', scale_units='xy', scale=1.0)
    
    ax.set_xlim(-1.4, 1.4)
    ax.set_ylim(-1.2, 1.2)
    
    ax.set_xlabel("%s axis" % axis_labels[index_x])
    ax.set_ylabel("%s axis" % axis_labels[index_y])
    
    ax.grid()
    
    colors = ['g', 'b']
    
    for i, bucket in enumerate(buckets):
        ax.quiverkey(q, centers[index_x,i], 1.2 * centers[index_y,i], 0, '%i' % bucket, 
                     coordinates='data', color=colors[0 if centers[index_z, i] > 0.0 else 1])
  
    if not prefix:
      prefix = 'centers' 
    
    save_fig(fig, prefix, dump=rp.get('dump'))
        
    return ax

# 
# Untested  
# 

def plot_spike(rp_set, fig, **kwargs):

  prefix = 'fuel'

  rp_main = rp_set.ray_profiles[0]

  limits = kwargs.get('limits', (rp_main.get('radinner'), rp_main.get('radouter')))

  y = rp_main.get('y')
  half_dx = 0.5 * (y[1] - y[0])

  spikes = kwargs.get('spikes', [])
  dumps = kwargs.get('dumps', [])
  times = kwargs.get('times', [])

  for i, rp in enumerate(rp_set):

    dump = rp.get('dump')
    time = rp.get('time')

    print dump,

    if dump in dumps or dump < 1:
      continue

    y = rp.get('y')
    y_limit = (y < limits[1]) & (y > limits[0])
    y = y[y_limit]

    fv = rp.get('fv')[0, y_limit]
    fv1 = 1.0 - fv

    rhobub = rp.get('rhobubble')[0, y_limit]
    rhospike = rp.get('rhospike')[0, y_limit]

    vols = (4. * np.pi / 3.) * ((y + half_dx)**3 - (y - half_dx)**3)
    bub_mass = vols * rhobub * fv1
    spike_mass = vols * rhospike * fv

    spikes.append(spike_mass.sum())
    dumps.append(dump)
    times.append(rp.get('time'))

  import matplotlib
  y_formatter = matplotlib.ticker.ScalarFormatter(useOffset=False)

  ax = fig.add_subplot(111)
  ax.set_title('H-quantity')
  ax.plot(dumps, spikes, 'k')

  ax.set_xlabel("Time")
  ax.yaxis.set_major_formatter(y_formatter)
  ax.grid(True)

  save_fig(fig, prefix, **kwargs)

  return dumps, times, spikes


def get_interface_location(rp, levels):

  nbuckets = rp.get('nbuckets')
  level_ct = len(levels)

  j_hi = rp.get_table('j_hi')
  y_hi = rp.get_table('y_hi')
  fv_hi = rp.get_table('fv_hi')[0,:,:]

  data = np.zeros((level_ct, nbuckets+1))

  #TODO: Don't loop
  for bucket in range(nbuckets+1):

    one_thing = fv_hi[:, bucket]

    first = np.argmax(one_thing > 0.0)
    last = np.argmin(one_thing < 1.0)

    j_restricted = j_hi[first-1:last-1]
    y_restricted = y_hi[first-1:last-1]

    fv_restricted = fv_hi[first-1:last-1, bucket]

    if fv_restricted.size == 0:
      continue

    for l, val in enumerate(levels):
      data[l, bucket] = np.interp(val, fv_restricted, y_restricted)

  return data

def get_interface_statistics(dump_data, dumps=None, results=None):
  log.info("get_interface_statistics")

  data_len = len(dump_data)

  if data_len == 0:
    return

  dumps = np.zeros(data_len, dtype=np.int32) if not dumps else dumps
  results = np.zeros((2, dump_data[0][1].shape[0], data_len)) if not results else results

  for i, d in enumerate(dump_data):

    dumps[i] = d[0]
    r = d[1][:,1:]

    print d[0],

    for l in range(r.shape[0]):
      results[0, l, i] = r[l].mean()
      results[1, l, i] = np.std(r[l])

  return dumps, results

def plot_interface(dumps, time_data, results, levels, **kwargs):

  styles_len = len(styles)
  colors_len = len(colors)

  fig = kwargs.get('fig', new_figure())
  quiet = kwargs.get('quiet', False)

  x_limits = kwargs.get('xlimits')
  y_limits = kwargs.get('ylimits')
  sd_only = kwargs.get('sd_only', False)

  ax = fig.add_subplot(1,1,1)
  ax.ticklabel_format(useOffset=False)
  ax.grid(True)
  ax.set_xlabel("dump #")
  ax.set_ylabel("r [1000 km]")

  if not sd_only:
    ax.set_title("Location of FV interface", y=1.08)
    for i, v in enumerate(levels):
      avg = results[0,i,:]
      sd = results[1,i,:]
      color = colors[ i % colors_len]

      ax.errorbar(dumps, avg, yerr=sd, label='fv=%f' % v,
                  fmt='k.-', ecolor=color)
  else:
    ax.set_title("Standard-Deviation of FV interface", y=1.08)
    for i, v in enumerate(levels):
      sd = results[1,i,:]
      style = styles[ i % styles_len]
      ax.plot(dumps, sd, style, label='fv=%f' % v)

  if x_limits:
    print "X LIMITS", x_limits
    ax.set_xlim(x_limits)
    #ax2.set_xlim(x_limits)

  if y_limits:
    ax.set_ylim(y_limits)
    #ax2.set_ylim(y_limits)

  ax.legend(loc=2)
  ax2 = ax.twiny()
  ax2.set_xlabel("time [min]")
  ax2.set_xticks(ax.get_xticks())
  ax2.set_xticklabels(time_data)
  ax2.set_xbound(ax.get_xbound())
  ax2.set_ybound(ax.get_ybound())


  save_fig(fig, 'interface', **kwargs)

  if quiet:
    matplotlib.pyplot.close(fig)

def plot_interface_analysis(rps, **kwargs):

  dumps = kwargs.get('dumps')
  time_data = kwargs.get('timing', [])
  prefix = kwargs.get('prefix')
  xlimits = kwargs.get('xlimits')
  levels = kwargs.get("levels", (0.5, 0.1, 0.01, 0.001, 1e-5))

  dump_data = []

  for rp in rps:
    dump = rp.get('dump')

    if dumps and dump not in dumps:
      continue

    print dump,

    d = get_interface_location(rp, levels)
    dump_data.append((dump, d))
    time_data.append('%0.f' % rp.get('timerescaled'))

  dumps, results = get_interface_statistics(dump_data)

  plot_interface(dumps, time_data, results, levels, filename='%s.png' % prefix,
                 xlimits=xlimits)

  plot_interface(dumps, time_data, results, levels, filename='%s_sd_only.png' % prefix,
                 xlimits=xlimits, sd_only=True)

  # Return dump_data so we can cache it
  return dump_data




#!/usr/bin/env python

"""
Plot the standard set of plots

Usage:

.. code:: bash

  plot_rprofiles.py [<rprofile_path>] [<start_dump>]
"""

import matplotlib as mpl
mpl.use("Agg")
#mpl.use("Cairo")
import matplotlib.pyplot as plt

#from matplotlib import rc
#rc('mathtext', default='regular')

import numpy as np
import time
import os
import subprocess
import sys

import lcse
from lcse_tools import ppm_plots

def main():

  if len(sys.argv) < 2:
    print "Usage:\n plot_rprofiles.py [<rprofile_path>] [<start_dump>]"
    return

  rprofile_path = sys.argv[1] if len(sys.argv) > 1 else './RProfile-01'
  start_dump = int(sys.argv[2]) if len(sys.argv) > 2 else 0

  plot_path = 'plots'
  sleep_time = 15 * 60

  fig = plt.figure()

  if not os.path.exists('plots'):
    os.makedirs('plots')

  try:
    # Load all the dumps from last time
    f = open('processed.dat', 'r')
    processed = [int(k) for k in f.readlines()]
    f.close()
  except IOError:
    processed = []

  print "Loading RProfiles from %s" % rprofile_path
  rp_set = lcse.rprofile_set(rprofile_path, lazy=False, logging=True, stride=1, start_dump=start_dump)

  if not rp_set.dumps:
    print "No RProfiles found"

#  volumes = rp_set.get_dump().get_cell_volumes()

  # Cache for the plot_spike plot
  fuel_dumps = []
  fuel_times = []
  fuel_spikes = []

  interface_data = []
  time_data = []
  energy_data = []

  levels = (0.5, 0.1, 0.01, 0.001, 1e-5)

  while True:
    rp_set.check_for_new(path=rprofile_path)

    for rp in rp_set:
      dump = rp.get('dump')

      if dump == 0 or dump in processed:
        print 'Skipping dump %04d' % dump
        continue

      print 'Processing dump %04d' % dump

      fig.clear()
      ppm_plots.plot_ceul_mach_global(rp, fig, path=plot_path)

      fig.clear()
      ppm_plots.plot_ekr_ekt_entropy(rp, fig, path=plot_path)

      # Linear plots
      for var in ["enuc"]:
      #for var in ["enuc", "fnuc"]:
        fig.clear()
        ppm_plots.plot_var_one_bucket(rp, var, fig, path=plot_path)

      # Log plots
      for var in ["enuc", "fv_hi"]:
      #for var in ["enuc", "fnuc", "fv_hi"]:
        fig.clear()
        ppm_plots.plot_log_var_one_bucket(rp, var, fig, path=plot_path)


      # Process per-dump things
      processed.append(dump)

      # Fuel interface
      d = ppm_plots.get_interface_location(rp, levels)
      interface_data.append((dump, d))

      time_data.append('%0.f' % rp.get("timerescaled"))

#      enuc = rp.get_table('enuc')[0,:,1:]
#      enuc *= volumes
#      energy_data.append(enuc.sum())

    # Routines that require all buckets
    fig.clear()
#    fuel_dumps, fuel_times, fuel_spikes = ppm_plots.plot_spike(rp_set, fig, dumps=fuel_dumps, times=fuel_times,
#                                                               spikes=fuel_spikes, path=plot_path)

#    dumps, results = ppm_plots.get_interface_statistics(interface_data)

#    xlimits = None
#    ppm_plots.plot_interface(dumps, time_data, results, levels, filename='fv_interface.png',
#                             xlimits=xlimits, path=plot_path)
#
#    ppm_plots.plot_interface(dumps, time_data, results, levels, filename='fv_interface_sd_only.png',
#                             xlimits=xlimits, sd_only=True, path=plot_path)

#    ax = fig.add_subplot(1,1,1)
#    ax.ticklabel_format(useOffset=False)
#    ax.grid(True)
#    ax.set_ylabel("energy")
#    ax.set_xlabel("time [min]")
#
#    ax.plot(time_data, np.log10(energy_data), label="enuc_total")
#    ax.set_ylim([-6,0])
#    ppm_plots.save_fig(fig, "enuc_total", dump=0000, path=plot_path)

    # Save processed dumps
    processed.sort()
    f = open('processed.dat', 'w')
    f.write('\n'.join(str(k) for k in processed))
    f.close()

    # Sleep for a while
    time.sleep(sleep_time)

if __name__ == '__main__':
  main()



#!/usr/bin/env python

'''
This thing consists of a master rank which sends out work to worker ranks. Both
master and worker do work using multiple python threads. These are not exactly OS
threads and can not really run on multiple cores but they are ok for data processing
since the executables will run on separate cores ok.

It can operate in 2 modes: Processing or Rendering.

In processing mode each dump-variable is reformated, bob2hv-ed, rendered to png
using the provided keys. Rendering is done as a step of the processing, not in
a separate thread. This makes things much more simple at the cost of less
efficient hardware usage. At some point we can probably overlap processing and
rendering and gain a minute or something processing time per dump. Too complicated
right now.

To use `python on Blue Waters <https://bluewaters.ncsa.illinois.edu/python>`_ one must load the gnu environment, 
load python, python mpi and opengl 

.. code:: bash

  module swap PrgEnv-cray PrgEnv-gnu
  module load bwpy
  module load bwpy-mpi
  module load opengl

'''

__version__ = 5.1

# TODO: Capture Ctrl+C  (nompi) http://snakesthatbite.blogspot.com/2010/09/cpython-threading-interrupting.html
# TODO: Make default options process everything
# TODO: Simplify command line options (--vars)
# TODO: Processing status report (dump: processed, hvs)

import argparse
import gzip
import hashlib
import io
import os
import threading
import time
import logging
import lockfile
import shutil
import signal
import subprocess
import sys
import tempfile
import Queue

log = logging.getLogger(__name__)
log.propagate = False
ch = logging.StreamHandler()
log.addHandler(ch)
log.setLevel(logging.INFO)
#log.setLevel(logging.DEBUG)

try:
  import lcse
  hvr_enabled = True
except ImportError:
  hvr_enabled = False
  error = "Warning: LCSE Python library not found, HVR rendering disabled"
  log.error(error)

from lcse_tools import ppm_plots

tag_ready = 1
tag_dump = 2

variable_list = [
    dict(name='fv', original='FV-hires-01', dir='FV-hiret-01', file='FV-hiret01', ext='bob8aaa', resolution=2),
    dict(name='vort', original='Lg10Vort-01', dir='Lg10Voru-01'),
    dict(name='enuc', original='Lg10ENUCbyP', dir='Lg10ENVCbyP'),
    dict(name='uy', original='TanhUY--001', dir='TanhUY-0001'),
    dict(name='divu', original='TanhDivU-01', dir='TanhDivV-01')
    ]

var_map = dict((k['name'], k) for k in variable_list)
#var_map = dict((k['dir'], k) for k in variable_list)
variables_original = [k['original'] for k in variable_list]

dir_t = '{dir:s}/{dump:04d}/{var:s}'
dir_var_t = '{dir:s}/{var:s}'
dir_var2_t = '{dir:s}/{var:s}-{suffix:s}'

filename_t = '{var:s}-{dump:04d}.{ext:s}'
path_t = '{dir:s}/{var:s}/{var:s}-{dump:04d}.{ext:s}'
path2_t = '{dir:s}/{var:s}-{suffix:s}/{var:s}-{suffix:s}-{dump:04d}.{ext:s}'

# 1010/FV-hiret-01/FV-hiret01-1010.bob8aaa
# 1010/Lg10Voru-01/Lg10Voru-01-1010.bobaaa
# 1010/TanhUY-0001/TanhUY-0001-1010.bobaaa
# 1010/TanhDivV-01/TanhDivV-01-1010.bobaaa
# 1010/Lg10ENVCbyP/Lg10ENVCbyP-1010.bobaaa

#TODO: Remove this garbage
class keyfile_reader(object):

  #TODO: Change image with / height
  def __init__(self, filename, hv_dir, img_dir, image_size):

    self.key_name = os.path.basename(filename)[:-4]

    self.hv_dir = os.path.normpath(hv_dir) if hv_dir else None
    self.img_dir = os.path.normpath(img_dir) if img_dir else None
    self.image_size = image_size

    self.key = lcse.hvr_key(filename)
    hvfile = self.key.get('hvfile')

    self.var = os.path.basename(hvfile)[:-8]

    if hv_dir:
      self.key._data['hvfile'] = os.path.join(hv_dir, os.path.basename(hvfile))

last_sigint = None

def sig_int_handler(signal, frame):
  global last_sigint

  log.info("Received signal %s last %s" % (signal, last_sigint))

  t = time.time()

  if last_sigint is not None and (t - last_sigint) < 5.0:
    log.info("Received second Ctrl+C within 5 seconds, exiting")
  sys.exit(0)

  last_sigint = t

def setup_logging(args):

  if not args.mpi:
    try:
      from colorlog import ColoredFormatter

      formatter = ColoredFormatter("%(log_color)s %(asctime)s - %(levelname)s : %(white)s%(message)s",
                                   datefmt=None,
                                   reset=True,
                                   log_colors={
                                     'DEBUG':    'cyan',
                                     'INFO':     'green',
                                     'WARNING':  'yellow',
                                     'ERROR':    'red',
                                     'CRITICAL': 'red,bg_white',
                                   },
                                   secondary_log_colors={},
                                   style='%'
                                  )
      ch.setFormatter(formatter)
      return
    except:
      pass

 # formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
  formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
  ch.setFormatter(formatter)

  if args.log:
    if not os.path.exists(args.logs_dir):
      os.makedirs(args.logs_dir)

    fh = logging.FileHandler(os.path.join(args.logs_dir, args.log))
    fh.setFormatter(formatter)
    log.addHandler(fh)

def execute(rank, thread, cmd, cwd=None):

  if cmd[0].startswith('.'):
    cmd[0] =  os.path.abspath(cmd[0])
    print cmd[0]

  cmd_line = ' '.join(cmd)
  log.info("rank %i.%i : exec '%s'" % (rank, thread, cmd_line))
  a = time.time()

  try:
    p = subprocess.Popen(cmd_line, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd, shell=True)
    p.communicate()
    ret_val = p.wait()
  except Exception as e:
    log.warning("execute() caught an exception %s" % e)
    ret_val = -1

  b = time.time()
  duration = b - a
  log.info("rank %i.%i : exec '%s' finished with code %i in %f.2" % (rank, thread, cmd_line, ret_val, duration))

  return (ret_val, duration)

def get_dumps(args):

  if args.auto:
    dumps_contents = [d for d in os.listdir(args.dumps_dir) if os.path.isdir(os.path.join(args.dumps_dir, d))]
    dumps = [int(c) for c in dumps_contents if c.isdigit()]

    if not dumps and args.render:
      dumps = list(set([int(hv[-7:-3]) for hv in os.listdir(args.hvs_dir) if hv.endswith('.hv')]))

  else:
    # Should be able to run for 1 or a range of dumps
    dumps = args.dumps if len(args.dumps) == 1 else range(args.dumps[0], args.dumps[1]+1)

  dumps.sort()

  if args.start_dump:
    log.info("Starting from dump %i" % args.start_dump)
    dumps = [d for d in dumps if d > args.start_dump]

  log.debug("Dumps are: %s" % dumps)

  return dumps

def get_keys(args):
  return

def remove_processed(vars, dumps):

  #TODO: Write this list as a shell script
  # for creating paths
  hv_contents = os.listdir(args.hvs_dir)

  # Generate list of all files
  hvr_dict = dict(ext='hvr', var='', dump='')

  for var in vars:
    for dump in dumps:
      names.append(filename_t.format(dump=dump, var=var, ext='hvr'))

#
def archive_dump(args, thread_id, dump):
  """
  We take files like ./0340/FV-hires-01/FV-hires01-0340.bob8abd and compress them into
  ./dump-archive/FV-hires-01/FV-hires-01-0340.gz/FV-hires01-0340.bob8abd.gz
  """

  results = {}

  log.info("rank %i.%i : started archiving dump %i" % (args.rank, thread_id, dump))

  output_dir = args.archive_dir

  gzip_path_t = '{dir:s}/{var:s}/{var:s}-{dump:04d}.gz/'
  gzip_filename = '{filename:s}.gz'

  #unpack_command = 'gzip -d {gzip:s};\nmv {file:s} {dest:s}'

  for var in variables_original:

    var_dir = var_map[var]['dir']

    bob_dir = dir_t.format(dir=args.dumps_dir, dump=dump, var=var_dir)
    bob_files = os.listdir(bob_dir)
    bob_files.sort()

    dir_out = gzip_path_t.format(dir=args.archive_dir, var=var_dir, dump=dump)

    if not os.path.exists(dir_out):
      os.makedirs(dir_out)

    hashes_path = os.path.join(dir_out, 'hashes.md5')
    moveum_path = os.path.join(dir_out, 'uncompressum.sh')
    verifyum_path = os.path.join(dir_out, 'verifyum.sh')

    if os.path.exists(hashes_path):
      log.warning("rank %i.%i : hashes file %s already exists skipping this dump/variable" % (args.rank, thread_id, hashes_path))
      continue

    hashes = []
    file_pairs = []

    for filename in bob_files:

      path_in = os.path.join(bob_dir, filename)
      filename_gz = gzip_filename.format(filename=filename)
      path_out = os.path.join(dir_out, filename_gz)

      file_pairs.append((filename, filename_gz))

      if os.path.exists(path_out):
#        log.info("rank %i.%i : gzip file %s exists, skipping" % (args.rank, thread_id, path_out))
        log.warn("rank %i.%i : gzip file %s exists, not skipping (for safety)" % (args.rank, thread_id, path_out))

      #TODO: Timing information
      with open(path_in, 'rb') as f_in, gzip.open(path_out, 'wb') as f_out:

        data_in = f_in.read()

        # Gzip save the data
        f_out.write(data_in)

        # Compute MD5 sum
        m = hashlib.md5()
        m.update(data_in)

        hash = m.hexdigest()
        hashes.append((filename, hash))

        log.info("rank %i.%i : gziping file %s (md5sum: %s) to %s" % (args.rank, thread_id, path_in,
                                                                      hash, path_out))

    if not hashes:
      continue

    # Write out the uncompressed hashes file
    with io.open(hashes_path, 'wb') as f:
      f.writelines('%s *%s\n' % (hash, filename) for filename, hash in hashes)
      log.debug("rank %i.%i : wrote hashes to %s" % (args.rank, thread_id, hashes_path))

    out_path = '../../uncompressed-dumps/%04d/%s' % (dump, var_dir)

    # Write out the moveum file
    with io.open(moveum_path, 'wb') as f:
      f.write("mkdir -p %s\n" % out_path)
      f.writelines('gzip -cd %s > %s/%s\n' % (filename_gz, out_path, filename) for filename, filename_gz in file_pairs)
      log.debug("rank %i.%i : wrote moveum to %s" % (args.rank, thread_id, hashes_path))

    # Write out hashes file
    with io.open(verifyum_path, 'wb') as f:
      for filename, hash in hashes:
        f.write('''a=`gzip -cd %s.gz | md5sum | awk '{print $1}'`; b="%s"; \n''' % (filename, hash))
        f.write('''if [ "$a" == "$b" ]; then echo "%s ok"; else echo %s fail"; fi; unset a; unset b; \n\n''' % (filename, filename));

      log.debug("rank %i.%i : wrote verifyum to %s" % (args.rank, thread_id, hashes_path))


  log.info("rank %i.%i : finished archiving dump %i" % (args.rank, thread_id, dump))

  return results

def process_dump(args, dump, thread_id):

  log.info("rank %i.%i : started processing dump %i" % (args.rank, thread_id, dump))

  results = {}

  # Process the dumps
  rank = args.rank

  vars = []
  existing = []

  to_process = []

  for var in args.vars:

    var_dir = var_map[var]['dir']

    hv_dir = dir_var_t.format(dir=args.hvs_dir, var=var_map[var].get('file', var_dir))
    hv_file = path_t.format(dir=args.hvs_dir, var=var_map[var].get('file', var_dir), dump=dump, ext='hv')

    if not os.path.exists(hv_file):
      vars.append(var)
      to_process.append((var, hv_dir))
    else:
      existing.append(var)

  # Check if we are duplicating something
  # TODO: Could reuse the paths here
  results['existing'] = existing

  xreformat_path = args.xreformat_exec
  bob2hv_path = args.bob2hv_exec

  # Clean up any stale old files
  rm_cmd = []
  if args.cleanup or args.reformat:
    rm_cmd.extend(['rm', '-rf'])
    rm_cmd.extend([dir_t.format(dir=args.dumps_dir, dump=dump, var=var_map[var]['dir']) for var in args.vars])

  if vars and args.reformat:
    dump_str = '%04i' % dump

    # Remove old and stale stuff before running reformat
    execute(args.rank, thread_id, rm_cmd)

    cwd = args.dumps_dir if args.multidir else dir_t.format(dir=args.dumps_dir, dump=dump, var=var_map[to_process[0][0]]['original'])
    log.warning("CWD is " + cwd)

    reformat_cmd = [xreformat_path, dump_str, dump_str, ">> ./logs/process.%s.log 2>&1" % dump_str]
    ret_val, t = execute(args.rank, thread_id, reformat_cmd, cwd=cwd)
    results['reformat'] = dict(ret=ret_val, duration=t)

  if args.bob2hv:
   results['bob2hv'] = {}
   results['mv_hv'] = {}

  for var, hv_dir in to_process:
    var_info = var_map[var]
    var_dir = var_info['dir']
    res = var_info.get('resolution', 1)
    ext = var_info.get('ext', 'bobaaa')
    file_prefix = var_info.get('file', var_dir)
    dump_dir = dir_t.format(dir=args.dumps_dir, dump=dump, var=var_dir)

    # Run bob2hv
    if args.bob2hv:
      if not os.path.exists(dump_dir):
        log.info("rank %i.%i : skipping %4i:%s because %s does not exist" % (args.rank, thread_id,
                                                                             dump, var, dump_dir))
        continue

      size = [str(res * i) for i in args.tile_size]
      teams = [str(t) for t in args.tiles]
      bob_filename = filename_t.format(dump=dump, var=file_prefix, ext=ext)
      bob2hv_cmd = [bob2hv_path, str(size[0]), str(size[1]), str(size[2]),
                    bob_filename, '-t', teams[0], teams[1], teams[2], '-s', '128',
                   ">> ../../logs/process.%04i.log 2>&1" % dump]
      ret_val, duration = execute(args.rank, thread_id, bob2hv_cmd, cwd=dump_dir)
      results['bob2hv']['vars'] = dict(ret=ret_val, duration=duration)

      if not os.path.exists(hv_dir):
         os.makedirs(hv_dir)

      hv_file_path = os.path.join(dump_dir, filename_t.format(dump=dump, var=file_prefix, ext='hv'))

      if not os.path.exists(hv_file_path):
        log.warning("rank %i.%i : hv %s not created for dump %4i" % (args.rank, thread_id,
                                                                     hv_file_path, dump))
        continue

      mv_cmd = ['mv', hv_file_path, hv_dir]

      ret_val, duration = execute(args.rank, thread_id, mv_cmd)
      results['mv_hv']['vars'] = dict(ret=ret_val, duration=duration)

      results['success'] = True

  if args.cleanup:
    execute(args.rank, thread_id, rm_cmd)

  log.info("rank %i.%i : finished processing dump %i" % (args.rank, thread_id, dump))

  return results

def render_dump(args, thread_id, dump):

  log.info("rank %i.%i : started rendering dump %4i" % (args.rank, thread_id, dump))

  hvr_path = args.hvr_exec
  keys = args.keys

  results = {'success': False}

  for key in keys:

    for var in args.specified_vars:

      if var not in args.lut_map:
        log.info("rank %i.%i : skipping dump %4i variable %s because no lut was provided" % (args.rank, thread_id, dump, var))
        continue

      size_str = 'x'.join(str(d) for d in key.key.get('imagesize'))

      hv_name = var_map[var].get('file', var_map[var]['dir'])
      hv_file = path_t.format(dir=args.hvs_dir, var=hv_name, dump=dump, ext='hv')

      if not os.path.exists(hv_file):
        log.warning("rank %i.%i : skipping dump %4i because %s does not exist" % (args.rank, thread_id,
                                                                                  dump, hv_file))
        continue

      var_name = var_map[var]['original']
      luts = args.lut_map[var]

      for lut in luts:

        #suffix = '%s-%s' % (lut.name, key.key_name)
        suffix = '%s-%s' % (lut.name, key.key_name) if len(luts) > 1 else key.key_name

        # Have to remap the annoying Fv-hires thing
        raw_file = path2_t.format(dir=args.images_dir, var=var_name, dump=dump, suffix=suffix, ext='raw')
        png_file = path2_t.format(dir=args.images_dir, var=var_name, dump=dump, suffix=suffix, ext='png')
        var_dir = dir_var2_t.format(dir=args.images_dir, var=var_name, suffix=suffix)

        if not os.path.exists(var_dir):
          os.makedirs(var_dir)

        if os.path.exists(png_file):
          log.info("rank %i.%i : skipping dump %4i because %s exists" % (args.rank, thread_id, dump,
                                                                       png_file))
#          results['success'] = True
          continue

        if not os.path.exists(raw_file):
          key.key._data['hvfile'] = hv_file
          key.key._data['imagefile'] = raw_file
          key.key.lut = lut

#        key_filename = os.path.join(args.temp_dir, "%s_tmp_%s.key" % (var, key.key_name))
#        key.key.save(key_filename)

          _, key_filename = tempfile.mkstemp(prefix='%s-%s' % (var_name, suffix), dir=args.temp_dir)

          with open(key_filename, 'wb') as key_file:
            key_file.write(str(key.key))

          log.debug('rank %i.%i : calling hvr for %s, temp %s' % (args.rank, thread_id, hv_file, key_filename))

          hvr_cmd = [hvr_path, '<', key_filename]
#          hvr_cmd = [hvr_path, '<', key_filename, ">> ./logs/render.%04i.log 2>&1" % dump]
          ret,_ = execute(args.rank, thread_id, hvr_cmd)

#          log.debug('rank %i.%i : calling hvr for %s' % (args.rank, thread_id, hv_file))
#          key_string = str(key)
#          p = subprocess.Popen(hvr_path, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
#          output = p.communicate(key_string)
#          ret = p.wait()
          log.debug('rank %i.%i : hvr exit with code %i' % (args.rank, thread_id, ret))

        if os.path.exists(raw_file):
          convert_cmd = ['convert','-depth','8','-flip', '-size', size_str, 'rgb:%s' % raw_file, png_file]
          ret,_ = execute(args.rank, thread_id, convert_cmd)

          # Delete .raw file only if we converted it to png successfuly
          if ret == 0:
            rm_raw_cmd = ['rm', '-f', raw_file]
            ret = execute(args.rank, thread_id, rm_raw_cmd)
            results['success'] = True

  log.info("rank %i.%i : finished rendering dump %4i, success %s" % (args.rank, thread_id, dump, results['success']))

  return results

def render_keyfile(args, thread_id, keyfile):

  hvr_path = args.hvr_exec

  key = keyfile_reader(keyfile, args.hvs_dir, args.images_dir, args.image_size)
  raw_file = os.path.join(args.images_dir, '%s.raw' % key.key_name)
  png_file = os.path.join(args.images_dir, '%s.png' % key.key_name)

  results = {}

  log.info("rank %i.%i : started rendering keyfile %s" % (args.rank, thread_id, keyfile))

  if os.path.exists(png_file):
    log.info("rank %i.%i : skipping keyfile %s because %s exists" % (args.rank, thread_id, keyfile, png_file))
    return results

  key.key.set('imagefile', raw_file)

  key_filename = "%s_%04i_%04i_tmp.key" % (key.key_name, args.rank, thread_id)
  key.save(key_filename)

  hvr_cmd = [hvr_path, '<', key_filename, ">> ./logs/render.%s.log 2>&1" % key.key_name]
  ret,_ = execute(args.rank, thread_id, hvr_cmd)

  if os.path.exists(raw_file):
    size = key.key.get('imagesize')
    convert_cmd = ['convert','-depth','8','-flip', '-size', 'x'.join(size), 'rgb:%s' % raw_file, png_file]
    ret, duration = execute(args.rank, thread_id, convert_cmd)

  if ret == 0:
    rm_raw_cmd = ['rm', '-f', raw_file]
    ret, _ = execute(args.rank, thread_id, rm_raw_cmd)

  log.info("rank %i.%i : finished rendering keyfile %s" % (args.rank, thread_id, keyfile))

  return results

def work_thread_main(args, in_queue, out_queue, thread_id):
  '''Process data from in_queue and write it to out_queue'''

# TODO: Make the lock more generic so it locks anything we need .lock.<work_type>
# TODO: The message payload should indicate the type of work we are doing

  processing = args.bob2hv or args.cleanup or args.reformat
  rendering = args.render and hvr_enabled

  log.info("rank %i.%i : work_thread_main" % (args.rank, thread_id))

  while True:
    work = in_queue.get()

    if isinstance(work, dict) and work.get('exit'):
      log.info("rank %i.%i : worker commanded to exit" % (args.rank, thread_id))
      break

    results = {'work':work}

    try:

      if not work and work != 0:
        log.info("rank %i.%i : worker is out of work, exiting. (work %s)" % (args.rank, thread_id, str(work)))
        break

      if args.plot:
        results.update(plot_dump(args, thread_id, work))

      if args.render_keys:
        keyfile = work
        results.update(render_keyfile(args, thread_id, keyfile))

      elif processing or rendering:
        dump = work

        log.debug("rank %i.%i : worker start processing unit %s" % (args.rank, thread_id, work))

        if processing:
          lock_path = dir_t.format(dir=args.dumps_dir, dump=dump, var='.dpp%s' % args.lock_name)
          lock = lockfile.LockFile(lock_path)

          if lock.is_locked():

            lock_age = time.time() - os.path.getctime(lock.lock_file)

            #  Anything over 60 minutes is stale
            if lock_age > 3600.0:
              log.info("rank %i.%i : found stale lock (%i s) on %4i" % (args.rank, thread_id,
                                                                        int(lock_age), dump))
              lock.break_lock()

          try:
            lock.acquire(timeout=-1)

            results.update(process_dump(args, dump, thread_id))

            lock.release()
          except lockfile.AlreadyLocked:
            log.warning("rank %i.%i : dump %4i is locked, skipping" % (args.rank, thread_id, dump))

        if rendering:
          results.update(render_dump(args, thread_id, work))

      elif args.archive:
        results.update(archive_dump(args, thread_id, work))

    except IndexError as e:
#    except Exception as e:
      log.error("rank %i.%i : failed processing work %s with Exception: %s" % (args.rank,
                                                                               thread_id, work,e))
    in_queue.task_done()
    out_queue.put(results)

def monitor_thread_main(args, work_queue, start_dumps):

  sleep_time = 60
  known = list(start_dumps)
  thread_id = 0

  while True:
    dumps = get_dumps(args)
    new_dumps = [d for d in dumps if d not in known]
    new_dumps.sort()
    # Last dumps is sort of invalid anyway
    new_dumps = new_dumps[:-1]

    if new_dumps:
      log.info("rank %i.%i : monitor found new dumps %s" % (args.rank, thread_id, new_dumps))

      for d in new_dumps:
        work_queue.put(d)
        known.append(d)

    time.sleep(sleep_time)

def plot_dump(args, thread_id, dump):
  """ Plot all of the things """

  rp = args.rp_set.get_file(dump)

  fig = args.fig
  plot_path = args.plot_dir

  dump = rp.get('dump')
  log.info("rank %i.%i : plotting dump %04i" % (args.rank, thread_id, dump))

  fig.clear()
  ppm_plots.plot_ceul_mach_global(rp, fig, path=plot_path)

  fig.clear()
  ppm_plots.plot_ekr_ekt_entropy(rp, fig, path=plot_path)

  # Linear plots
  for var in ["enuc", "fnuc"]:
    fig.clear()
    ppm_plots.plot_var_one_bucket(rp, var, fig, path=plot_path)

  # Log plots
  for var in ["enuc", "fnuc", "fv_hi"]:
    fig.clear()
    ppm_plots.plot_log_var_one_bucket(rp, var, fig, path=plot_path)

  # Process per-dump things
  for bucket in range(1, rp.get('nbuckets') + 1):
    for var in ["enuc", "fv_hi", "p"]:

      fig.clear()
      ppm_plots.plot_log_var_one_bucket(rp, var, fig, path=plot_path, bucket=bucket)

      fig.clear()
      ppm_plots.plot_var_one_bucket(rp, var, fig, path=plot_path, bucket=bucket)

  return {}

def worker_thread_manager(args, in_queue, out_queue):

  threads = []
  nthreads = args.threads

  for thread_id in range(1, nthreads+1):
    log.info("rank %i.%i : worker_thread_manager spawning thread %i " % (args.rank, 0, thread_id))

    t = threading.Thread(target=work_thread_main, args=(args, in_queue, out_queue, thread_id))
    t.daemon = True
    t.start()

    threads.append(t)

  return threads

def mpi_worker_main(comm, args):
  '''Query server for new data and process it'''

  processing = args.bob2hv

  process_q = Queue.Queue()
  ready_q = Queue.Queue()

  nthreads = args.threads

  log.debug("rank %i.%i : mpi_worker_main asking for work" % (args.rank, 0))

  comm.send({'ready':True, 'units':nthreads}, dest=0, tag=1)
  data = comm.recv(source=0, tag=2)
  work_units = data.get('work')

  threads = worker_thread_manager(args, process_q, ready_q)

  while work_units:

    for w in work_units:
      process_q.put(w)

    log.info("rank %i.%i : Waiting for item" % (args.rank, 0))
    results = ready_q.get()
    log.info("rank %i.%i : Got item %s " % (args.rank, 0, results))

    comm.send({'ready':True, 'results':results}, dest=0, tag=1)
    log.info("rank %i.%i : sent %s to leader" % (args.rank, 0, results))

    data = comm.recv(source=0, tag=2)
    log.info("rank %i.%i : recv %s from leader" % (args.rank, 0, data))
    work_units = data.get('work')

  # Wait for all items to finish
  log.info("rank %i.%i : mpi_worker_main waiting for all items to finish" % (args.rank, 0))
  process_q.join()

  for i in range(nthreads):
    process_q.put({'exit':True})

  results = []

  log.info("rank %i.%i : mpi_worker_main collecting final results" % (args.rank, 0))
  while not process_q.empty():
    results.append(process_q.get())

  log.info("rank %i.%i : mpi_worker_main reporting final results" % (args.rank, 0))
  comm.send({'ready':False, 'results':results}, dest=0, tag=1)

def run_el_presidente(args):
  '''Main routine for work distribution rank'''

  rendering = (args.render_keys or args.render) and hvr_enabled

  log.info("El Presidente version %s: starting %s " % (str(__version__), "rendering" if (args.render_keys) else "processing"))

  # Make directories
  dirs_to_make = [args.logs_dir]

  if args.bob2hv: dirs_to_make.append(args.hvs_dir)
  if rendering: dirs_to_make.extend([args.images_dir])
  if args.plot: dirs_to_make.append(args.plots_dir)

  for d in dirs_to_make:
    if not os.path.exists(d):
      os.makedirs(d)

  process_q = Queue.Queue()
  ready_q = Queue.Queue()

  work_units = []
  processed = []

  if args.render_keys:
    keys_dir = args.render_keys
    work_units = [dict(keyfile=os.path.join(keys_dir, f)) for f in os.listdir(keys_dir) if f.endswith('.key')]
  else:
    work_units = get_dumps(args)

  for w in work_units:
    process_q.put(w)

  log.info("El Presidente: starting worker-thread manager")
  threads = worker_thread_manager(args, process_q, ready_q)

  if args.monitor:
    #TODO: This should be able to monitor a directory for keys?
    t = threading.Thread(target=monitor_thread_main, args=(args, process_q, work_units))
    t.daemon = True
    t.start()

  if args.mpi:
    # Start MPI things that will be handling workers

    from mpi4py import MPI

    comm = MPI.COMM_WORLD
    size = comm.Get_size()
    ranks = range(1,size)
    status = MPI.Status()

    wait_q = Queue.Queue()

    while True:
      msg_in = comm.recv(source=MPI.ANY_SOURCE, status=status, tag=tag_ready)

      log.debug("El Presidente: Received message %s" % msg_in)

      if status is None:
         log.error("el_presidente: Received status None")
         continue

      source = status.Get_source()
      units = msg_in.get('units', 1)

      log.info("El Presidente: Recieved request for %i units for %i " % (units, source))

      msg_out = {}
      work_out = []

      # Here if we are a monitor we will block on this and wait indefinitely for more
      # queue elements. If we are not a monitor we should terminate when the queue
      # is empty.

      try:
        for i in range(units):
          work_out.append(process_q.get(args.monitor))
      except Queue.Empty:
        log.info("El Presidente: out of work")
        break

      log.info("El Presidente: sending %s work to %i " % (work_out, source))
      comm.send({'work':work_out}, dest=source, tag=tag_dump)

    # Clean up
    log.info("Cleaning up %s " % ranks)
    for source in ranks:
      log.info("Cleaning up rank %i" % source)
      msg_in = comm.recv(source=source, status=status, tag=tag_ready)
      comm.send({'work':[]}, dest=source, tag=tag_dump)

  log.info("El Presidente: Waiting for Processing Queue to be complete")
  process_q.join()

  log.info("El Presidente: Sending Quit messages")
  for i in range(args.threads):
    process_q.put({'exit':True})

  log.info("El Presidente: Exiting")
  
  
def build_argparse():

  parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

  parser.add_argument('--dumps', nargs='+',type=int, help='Manually specify dump number or dump range (e.g. `--dumps 355` or `--dumps 800 900)`')
  parser.add_argument('--start-dump', type=int, help='Start processing from this dump')
  parser.add_argument('--threads', type=int, default=1, help='Number of threads')
  parser.add_argument('--local', dest='mpi', action='store_false', help='Run without MPI')
  parser.add_argument('--process', action='store_true', help='reformat, generate HVs, move, cleanup')
  parser.add_argument('--render', action='store_true', help='Render dump after processing')
  parser.add_argument('--archive', action='store_true', help='Create a gzip archive of the dumps')
  parser.add_argument('--plot', action='store_true', help='Plot RProfiles')
  parser.add_argument('--plot-dir', default='./plots', help='')

  parser.add_argument('--archive-dir', default='./dump-archive', help='Directory containing dump files')
  parser.add_argument('--dumps-dir', default='.', help='Directory containing dump files')
  parser.add_argument('--hvs-dir', default='./hvs', help='Directory to move HV files')
  parser.add_argument('--logs-dir', default='./logs', help='Logs directory')
  parser.add_argument('--plots-dir', default='./plots', help='Directory for plots')
  parser.add_argument('--temp-dir', help='Temp directory')

  parser.add_argument('--lock-name', default='', help='Hack: Unique name for this lock (e.g. "bob" or "bob8")')
  parser.add_argument('--log', help='Log filename')

  p_group = parser.add_argument_group('Processing Options')
  p_group.add_argument('--reformat', action='store_true', help='Reformat bobs')
  p_group.add_argument('--bob2hv', action='store_true', help='Generate HV files')
  p_group.add_argument('--cleanup', action='store_true', help='Remove reformated bobs')
  p_group.add_argument('--monitor', action='store_true', help='Keep checking for new dumps to appear')
  p_group.add_argument('-t', '--tiles', nargs='+', type=int, help='Number of tiles (1 or 3 numbers)')
  p_group.add_argument('-g', '--tile-size', nargs='+', type=int, help='Size of a single tile (1 or 3 numbers)')
  p_group.add_argument('--bob2hv-exec', default='/u/sciteam/sandalsk/tools/bob2hv', help='')
  p_group.add_argument('--hvr-exec', default='/u/sciteam/sandalsk/lcse_serv/hvr_cl.exe', help='')
  p_group.add_argument('--xreformat-exec', default='./xreformat', help='')
  p_group.add_argument('--no-multidir', dest='multidir', action='store_false', default=True)

  r_group = parser.add_argument_group('Rendering Options')
  r_group.add_argument('--image-size', nargs='+', type=int, help='Image size as two space separated numbers (e.g. 1280 768). By default it uses the information from they key file')
  r_group.add_argument('--render-keys', help='Render key files in the given directory')
  r_group.add_argument('--keys', dest='key_files', nargs='+', help='A directory of key files or a list of key files to render for the selected dumps')
  r_group.add_argument('--images-dir', default='./images', help='Directory to store images')

  # Variable specific color maps that are used instead of the stuff inside the .lut
  r_group.add_argument('--lut-fv', nargs='+', help='')
  r_group.add_argument('--lut-vort', nargs='+', help='')
  r_group.add_argument('--lut-uy', nargs='+', help='')
  r_group.add_argument('--lut-enuc', nargs='+', help='')
  r_group.add_argument('--lut-divu', nargs='+', help='')

  v_group = parser.add_argument_group('Variables')
  v_group.add_argument('--all-vars', action='store_true', help='Process all things (fv, enuc, vort, divu, uy) (default)')
  v_group.add_argument('--fv', action='append_const', const='fv', dest='vars', help='FV High-Res (FV-hires-01)')
  v_group.add_argument('--vort', action='append_const', const='vort', dest='vars', help='Vorticity (Lg10Vort-01)')
  v_group.add_argument('--uy', action='append_const', const='uy', dest='vars', help='(TanhUY--001)')
  v_group.add_argument('--enuc', action='append_const',  const='enuc', dest='vars', help='Process Enuc files (Lg10ENUCbyP)')
  v_group.add_argument('--divu', action='append_const', const='divu', dest='vars', help='(TanhDivU-01')

#  v_group.add_argument('--fv', action='append_const', const='FV-hiret-01', dest='vars', help='FV High-Res (FV-hires-01)')
#  v_group.add_argument('--vort', action='append_const', const='Lg10Voru-01', dest='vars', help='Vorticity (Lg10Vort-01)')
#  v_group.add_argument('--uy', action='append_const', const='TanhUY-0001', dest='vars', help='(TanhUY--001)')
#  v_group.add_argument('--enuc', action='append_const',  const='Lg10ENVCbyP', dest='vars', help='Process Enuc files (Lg10ENUCbyP)')
#  v_group.add_argument('--divu', action='append_const', const='TanhDivV-01', dest='vars', help='(TanhDivU-01')

  return parser

def main():


  # Available variables
  parser = build_argparse()
  
  args = parser.parse_args()

  setup_logging(args)

  if not args.render_keys:
    args.auto = not (args.dumps and len(args.dumps) <= 2)

  args.rank = 0


  # If the grid is symmetric we want to be able to specify only one size
  if args.tile_size:
    args.tile_size = args.tile_size if len(args.tile_size) == 3 else 3 * [args.tile_size[0]]

  if args.tiles:
    args.tiles = args.tiles if len(args.tiles) == 3 else 3 * [args.tiles[0]]

  if args.process:
    args.reformat = True
    args.bob2hv = True
    args.cleanup = True

  if args.plot:
    args.rp_set = lcse.rprofile_set(path="./RProfile-01", lazy=False, logging=True)
    args.fig = ppm_plots.new_figure()

  if args.vars:
    args.specified_vars = args.vars
  else:
    args.vars = var_map.keys()
    args.specified_vars = []

  if args.bob2hv and not (args.tile_size and args.tiles):

    reformat_file = 'Reformat-mPPM.F'

    if os.path.exists(reformat_file):
      defines = dict(l.split()[1:] for l in open(reformat_file).readlines() if l.startswith('#define'))
      defines_expanded = dict([(k, defines.get(v, v)) for k,v in defines.items()])

      args.tiles = [int(defines_expanded[v]) for v in ['nnxteams', 'nnyteams', 'nnzteams']]
      args.tile_size = [int(defines_expanded['nnnn%s' % v]) * int(defines_expanded['nnt%sbricks' % v]) for v in ['x', 'y', 'z']]

      log.info("Parsing %s indicates %s tiles of size %s" % (reformat_file, args.tiles, args.tile_size))
    else:
      log.error("Tile-size and number of tiles are required for bob2hv. Specify -t and -g options or a `Reformat-mPPM.F` file.")
      return 1

  # Here we go over the key files and create key objects
  args.keys = []

  if args.key_files:

    if len(args.key_files) == 1 and os.path.isdir(args.key_files[0]):
      keys_dir = args.key_files[0]
      args.key_files = [os.path.join(keys_dir, f) for f in os.listdir(keys_dir) if f.endswith('.key')]

    for k in args.key_files:
      log.debug("main() generating key for %s" % (k))
      key = keyfile_reader(k, args.hvs_dir, args.images_dir, args.image_size)
      args.keys.append(key)

  lut_file_map = {}

  # HACK:
  if args.lut_fv:
    lut_file_map['fv'] = args.lut_fv
  if args.lut_vort:
    lut_file_map['vort'] = args.lut_vort
  if args.lut_enuc:
    lut_file_map['enuc'] = args.lut_enuc
  if args.lut_divu:
    lut_file_map['divu'] = args.lut_divu
  if args.lut_uy:
    lut_file_map['uy'] = args.lut_uy

  args.lut_map = {}

  if lut_file_map:
    for k, v in lut_file_map.items():
      args.lut_map[k] = [lcse.lut(filename=f) for f in v]

  # Register the signal handler
#  signal.signal(signal.SIGINT, sig_int_handler)
#  signal.pause()

  if args.mpi:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD

    args.rank = comm.Get_rank()
    size = comm.Get_size()
    args.mpi = size > 0

    log.info("We are running in MPI mode, I am %s of %s" % (args.rank, size))

    if args.rank > 0:
      mpi_worker_main(comm, args)
      return

  # Either way we run El Presidente
  run_el_presidente(args)

if __name__ == '__main__':
  main()



#import os
from setuptools import setup

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
#def read(fname):
#    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "lcse_tools",
    version = "5.3",
    author = "Stou Sandalski",
    author_email = "sandalski@astro.umn.edu",
    description = ("Misc tools"),
    scripts = ('src/dpp_server.py', 'src/eval_pp.py', 'src/plot_rprofiles.py', 'src/check_missing.py',
               'src/pngs2movies.py', 'src/layer_images.py'),
    data_files = [('fonts', ['fonts/README.txt', 'fonts/LICENSE.txt', 'fonts/Roboto-Black.ttf',
                              'fonts/Roboto-Bold.ttf', 'fonts/Roboto-Light.ttf', 'fonts/Roboto-Medium.ttf',
                              'fonts/Roboto-Regular.ttf', 'fonts/Roboto-Thin.ttf'])],
    packages = ['lcse_tools'],
    package_dir = {'lcse_tools':'lib'},
#    package_data = {'':['fonts/*.txt', 'fonts/*.ttf']},
    install_requires = ['colorlog', 'numpy', 'lcse', 'lockfile'],
    license = "Apache 2.0",
    keywords = "utils",
    url = "",
    zip_safe=False,
#    long_description=read('README'),
#    classifiers=[
#        "Development Status :: 3 - Alpha",
#        "Topic :: Libraries",
#    ],
)
